import pandas as pd
import sparse
from scipy.sparse import coo_matrix
from pandas.api.types import CategoricalDtype


class SparseLodes:
    def __int__(self, df):
        geocodes = pd.unique(df[['w_geocode', 'h_geocode']].values.ravel('K'))
        jobs = [c for c in df.columns if c not in ['w_geocode', 'h_geocode']]
        shape = (len(geocodes), len(geocodes))

        # Map geocodes to coded index
        self.jobs_cat = CategoricalDtype(categories=sorted(jobs), ordered=True)
        self.geocodes_cat = CategoricalDtype(categories=sorted(geocodes), ordered=True)
        self.h_geocode_index = df['h_geocode'].astype(self.geocodes_cat).cat.codes
        self.w_geocode_index = df['w_geocode'].astype(self.geocodes_cat).cat.codes

        # Create dict of sparse matrices
        mat_dict = {k: coo_matrix((df[k], (self.w_geocode_index, self.h_geocode_index)), shape=shape) for k in jobs}

        # Stack into an n*n*k sparse coordinatate (COO) matrix
        # where n is the number of zones and k are the strata
        self.sparse_mat = sparse.stack([sparse.COO(m) for m in mat_dict.values()], axis=-1)




import fiona
import gzip
import os
import io
import geopandas as gpd
import pandas as pd
from lodes_star.utils import get_latest_year, get_file_list, fetch_bytes
from lodes_star.state_codes import State


def fetch_lodes(state,
                zone_types=['od', 'rac', 'wac'],
                job_types=['JT00'],
                segments=['S000'],
                year=None,
                cache=True):

    zone_types = [zone_types] if isinstance(zone_types, str) else zone_types
    job_types = [job_types] if isinstance(job_types, str) else job_types
    segments = [segments] if isinstance(segments, str) else segments
    base_url = 'https://lehd.ces.census.gov/data/lodes/LODES7'

    if not year:
        year = get_latest_year(base_url, state)
        print('No year specified, defaulting to latest year ' + year)

    # Create file list
    flist = get_file_list(base_url=base_url,
                          state=state,
                          zone_types=zone_types,
                          segments=segments,
                          job_types=job_types,
                          year=year)

    # Downloading files
    lodes = {}
    for fname, file_url in flist.items():
        # Fetch the file from URL or cache
        suffix = '{}/{}'.format(list(flist.keys()).index(fname) + 1, len(flist))
        bytes_data = fetch_bytes(file_url, suffix, cache)

        # Decompress the gzip bytes data and read into pandas dataframe
        string_io = io.StringIO(gzip.decompress(bytes_data).decode('utf-8'))
        df = pd.read_csv(string_io, dtype={'h_geocode': str, 'w_geocode': str, 'createdate': str})

        # Stash it
        key = fname.replace('.csv.gz', '')
        lodes[key] = df

    return lodes


# Fetch Census Blocks
def fetch_geo(state, geography, year='2021', cache=True):
    assert(len(year) == 4)
    year = str(year)

    if len(state) > 2:
        state = State.name2abb[state.capitalize()].lower()
    state_num = State.abb2code[state.upper()]

    altgeo = geography.lower()
    if geography == 'TABBLOCK':
        altgeo = geography.lower() + '10'

    # Format the file url
    url_template = 'https://www2.census.gov/geo/tiger/TIGER{year}/{geography}/tl_{year}_{fips}_{altgeo}.zip'
    # url_template = 'https://www2.census.gov/geo/tiger/TGRGDB{yy}/tlgdb_{yyyy}_a_{state_num}_{state}.gdb.zip'
    file_url = url_template.format(**{'year': year, 'geography': geography, 'fips': state_num, 'altgeo': altgeo})

    # Fetch the file from URL or cache
    bytes_data = fetch_bytes(file_url, cache=cache)

    if '.gdb' in os.path.basename(file_url):
        # Read geodatabase into geopandas
        with fiona.io.ZipMemoryFile(bytes_data) as zip_memory_file:
            with zip_memory_file.open(os.path.basename(file_url).rstrip('.zip')) as collection:
                geodf = gpd.GeoDataFrame.from_features(collection)
    else:
        geodf = gpd.read_file(io.BytesIO(bytes_data))

    return geodf



import requests
import io
import os
import glob
import shutil
from bs4 import BeautifulSoup
from tqdm.auto import tqdm
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
from urllib.parse import urlparse


###
def fetch_bytes(file_url, suffix="", cache=True):
    if isinstance(cache, str):
        file_path = os.path.join(cache, urlparse(file_url).path.lstrip('/'))
    else:
        file_path = os.path.join('./cache', urlparse(file_url).path.lstrip('/'))

    file_name = os.path.basename(file_url)

    if not os.path.exists(file_path.replace(file_name, '')):
        os.makedirs(file_path.replace(file_name, ''))

    # Check if already cached
    if not os.path.exists(file_path):
        s = requests.Session()
        response = requests_retry_session(session=s).get(file_url, stream=True, timeout=5)

        # file obj as real file or nulls
        obj = open(file_path, 'wb') if cache else open(os.devnull, 'wb')

        bytes_data = b''
        with tqdm.wrapattr(obj, "write",
                           desc=' '.join(['Fetching', file_name, suffix]),
                           total=int(response.headers.get("Content-Length", 0))
                           ) as out:

            # save the output to a file
            for buf in response.iter_content(io.DEFAULT_BUFFER_SIZE):
                out.write(buf)
                bytes_data += buf

        return bytes_data

    else:
        print(' '.join(['Loading cached', file_name, suffix]))
        return open(file_path, 'rb').read()


def get_cache_list(cache_dir=os.getcwd(), full_path=True):
    if full_path:
        return [f for f in glob.glob(os.path.join(cache_dir, '**/*.csv.gz'), recursive=True)]
    else:
        return [os.path.basename(f) for f in glob.glob(os.path.join(cache_dir, '**/*.csv.gz'), recursive=True)]


def get_file_list(base_url, state, zone_types, segments, job_types, year):
    file_list = {}

    for zone in zone_types:
        # Get list of relevant files
        url = '/'.join([base_url, state.lower(), zone.lower()])

        # HTTPS session
        s = requests.Session()
        response = requests_retry_session(session=s).get(url, stream=True, timeout=5)

        # Parse the hrefs from page
        hrefs = [x.get('href') for x in BeautifulSoup(response.text, 'html.parser').find_all('a')]

        for node in hrefs:
            is_jt = any([True for x in job_types if x in node])
            is_seg = any([True for x in segments if x in node])
            is_year_file = node.endswith('.csv.gz') and year in node

            is_odfile = zone.lower() == 'od' and is_year_file and is_jt
            is_racwacfile = zone.lower() != 'od' and is_year_file and is_jt and is_seg

            if is_odfile or is_racwacfile:
                file_list[node] = os.path.join(url, node)

    return file_list


def get_latest_year(base_url, state):
    common_years = None
    for zone in ['od', 'rac', 'wac']:
        url = '/'.join([base_url, state.lower(), zone.lower()])

        s = requests.Session()
        response = requests_retry_session(session=s).get(url, timeout=5)

        hrefs = [x.get('href') for x in BeautifulSoup(response.text, 'html.parser').find_all('a')]
        years = [node.replace('.csv.gz', '')[-4:] for node in hrefs if node.endswith('.csv.gz')]

        if not common_years:
            common_years = set(years)
        else:
            common_years = common_years.intersection(years)

    return max(common_years)


def requests_retry_session(
    retries=3,
    backoff_factor=0.3,
    status_forcelist=(500, 502, 504),
    session=None,
):
    session = session or requests.Session()
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    return session


def iterable_to_stream(iterable, buffer_size=io.DEFAULT_BUFFER_SIZE):
    """
    Lets you use an iterable (e.g. a generator) that yields bytestrings as a read-only
    input stream.

    The stream implements Python 3's newer I/O API (available in Python 2's io module).
    For efficiency, the stream is buffered.
    """
    class IterStream(io.RawIOBase):
        def __init__(self):
            self.leftover = None

        def readable(self):
            return True

        def readinto(self, b):
            try:
                l = len(b)  # We're supposed to return at most this much
                chunk = self.leftover or next(iterable)
                output, self.leftover = chunk[:l], chunk[l:]
                b[:len(output)] = output
                return len(output)
            except StopIteration:
                return 0    # indicate EOF
    return io.BufferedReader(IterStream(), buffer_size=buffer_size)


def stream_to_file(response, file_path, desc_lab=""):
    # check header to get content length, in bytes
    total_length = int(response.headers.get("Content-Length"))
    # implement progress bar via tqdm
    with tqdm.wrapattr(response.raw, "read", total=total_length, desc=desc_lab) as raw:
        # save the output to a file
        with open(file_path, 'wb') as output:
            shutil.copyfileobj(raw, output)



"""
lodes_star

For scraping Longitudinal Origin-Destination Employment Statisics (LODES) data from the US Census as well as fetching Census Block GIS shapefiles.
"""

__version__ = "0.2.5"
__author__ = 'Nicholas Fournier'



from setuptools import setup

setup(
    name='lodes_star',
    version='0.2.5',
    description='For scraping Longitudinal Origin-Destination Employment Statistics (LODES) data from'
                ' the US Census as well as fetching Census Block GIS shapefiles.',
    url='https://github.com/nick-fournier/lodes-star',
    author='Nicholas Fournier',
    author_email='nichfournier@gmail.com',
    license='GNU',
    packages=['lodes_star'],
    include_package_data=True,
    install_requires=['beautifulsoup4==4.11.1',
                      'Fiona==1.8.21',
                      'geopandas==0.10.2',
                      'matplotlib==3.5.2',
                      'pandas==1.4.2',
                      'requests==2.27.1',
                      'scipy==1.8.1',
                      'setuptools==60.2.0',
                      'sparse==0.13.0',
                      'tqdm==4.64.0',
                      'urllib3==1.26.9'],
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.9',
    ],
)



import csv
import os


class Geographies:
    with open(os.path.join(os.path.dirname(__file__), 'geographies.csv')) as csvfile:
        next(csvfile)
        csv_reader = csv.reader(csvfile)
        geographies = dict(csv_reader)

    @classmethod
    def list(cls):
        whitespace = max([len(x) for x in cls.geographies.keys()])
        for k, v in cls.geographies.items():
            print(k, ' '*(whitespace - len(k)), v)
        return

    def __new__(cls, geography=None):
        if geography in cls.geographies.keys():
            return cls.geographies[geography]
        else:
            print("Geography not found")


class State:
    with open(os.path.join(os.path.dirname(__file__), 'states.csv')) as csvfile:
        csv_reader = csv.DictReader(csvfile)
        states = list(csv_reader)

    abb2code = {x['abb']: x['code'] for x in states}
    code2abb = {v: k for k, v in abb2code.items()}

    name2code = {x['name']: x['code'] for x in states}
    code2name = {v: k for k, v in name2code.items()}

    name2abb = {x['name']: x['abb'] for x in states}
    abb2name = {v: k for k, v in name2abb.items()}
